@tool
extends Node3D

const SIZE = 64
const XZ_SCALE = 2.0
const STEEPNESS = 0.02

const CONFLICT_COLORS = [
	Color.RED,
	Color.YELLOW,
	Color.GREEN,
	Color.CYAN,
	Color.BLUE,
	Color.MAGENTA
]

@export var rebake: bool :
	set(_value): _rebake()

@export var merge_vertices: bool

var polygon_meshes: Array

func _ready() -> void:
	_rebake()

func _rebake() -> void:
	var nav_region := get_node_or_null(^"NavigationRegion3D")
	if !nav_region:
		nav_region = NavigationRegion3D.new()
		nav_region.name = &"NavigationRegion3D"
		add_child(nav_region, true)
		nav_region.owner = self
	
	var mesh_instance := get_node_or_null(^"MeshInstance3D")
	if !mesh_instance:
		mesh_instance = MeshInstance3D.new()
		mesh_instance.name = &"MeshInstance3D"
		add_child(mesh_instance, true)
		mesh_instance.owner = self
	
	var nav_mesh: NavigationMesh = nav_region.navigation_mesh if nav_region.navigation_mesh else NavigationMesh.new()
	
	var map := get_world_3d().navigation_map
	assert(NavigationServer3D.map_get_cell_size(map) == nav_mesh.cell_size)
	assert(NavigationServer3D.map_get_cell_height(map) == nav_mesh.cell_height)
	
	# Create a parabolic U-shape mesh
	var st := SurfaceTool.new()
	st.begin(Mesh.PRIMITIVE_TRIANGLES)
	st.set_uv(Vector2.ZERO)
	for x in range(-SIZE, SIZE):
		for z in range(-SIZE, SIZE):
			st.add_vertex(_vertex(x+0, z+0))
			st.add_vertex(_vertex(x+1, z+0))
			st.add_vertex(_vertex(x+0, z+1))
			st.add_vertex(_vertex(x+1, z+0))
			st.add_vertex(_vertex(x+1, z+1))
			st.add_vertex(_vertex(x+0, z+1))
	st.index()
	st.generate_normals()
	st.generate_tangents()
	mesh_instance.mesh = st.commit()
	
	# Bake a navmesh for the U-shaped mesh
	var source_geometry_data := NavigationMeshSourceGeometryData3D.new()
	source_geometry_data.add_mesh(mesh_instance.mesh, Transform3D.IDENTITY)
	NavigationMeshGenerator.bake_from_source_geometry_data(nav_mesh, source_geometry_data)
	
	if merge_vertices:
		_postprocess_nav_mesh(nav_mesh)
	_check_for_errors(nav_mesh)
	
	nav_region.navigation_mesh = null
	nav_region.navigation_mesh = nav_mesh

func _vertex(x: float, z: float) -> Vector3:
	x *= XZ_SCALE
	z *= XZ_SCALE
	return Vector3(x, _get_height(x, z), z)

func _get_height(x: float, z: float) -> float:
	return (x-z) * (x-z) * STEEPNESS

func _check_for_errors(nav_mesh: NavigationMesh) -> void:
	assert(nav_mesh != null)
	assert(nav_mesh.cell_size > 0.0)
	assert(nav_mesh.cell_height > 0.0)
	
	var cell_size: Vector3 = Vector3(nav_mesh.cell_size, nav_mesh.cell_height, nav_mesh.cell_size)
	var vertices: PackedVector3Array = nav_mesh.get_vertices()
	
	var edges := {} # edge key -> Array of { polygon_index: int, point_index: int, vertex1: Vector3, vertex2: Vector3 }
	
	for i in range(nav_mesh.get_polygon_count()):
		var polygon := nav_mesh.get_polygon(i)
		
		for j in range(polygon.size()):
			var vertex := vertices[polygon[j]]
			var next_vertex := vertices[polygon[(j + 1) % polygon.size()]]
			var edge_key := _get_edge_key(vertex, next_vertex, cell_size)
			if !edges.has(edge_key):
				edges[edge_key] = []
			edges[edge_key].push_back({
				polygon_index = i,
				point_index = j,
				vertex1 = vertex,
				vertex2 = next_vertex
			})
	
	for poly in polygon_meshes:
		remove_child(poly)
		poly.queue_free()
	polygon_meshes.clear()
	
	var color_index := 0
	var polygons := []
	for connections in edges.values():
		if connections.size() <= 2:
			continue
		
		print("Edge shared by more than 2 polygons:")
		for edge_data in connections:
			print("\tPolygon ", edge_data.polygon_index, " Edge: ", edge_data.vertex1, " -> ", edge_data.vertex2)
			if !polygons.has(edge_data.polygon_index):
				polygons.push_back(edge_data.polygon_index)
	
	for polygon_index in polygons:
		var mi := MeshInstance3D.new()
		mi.name = "Poly" + str(polygon_index)
		add_child(mi, true)
		var polygon := nav_mesh.get_polygon(polygon_index)
		var st := SurfaceTool.new()
		st.begin(Mesh.PRIMITIVE_TRIANGLE_STRIP)
		st.set_uv(Vector2.ZERO)
		for index in polygon:
			st.add_vertex(vertices[index])
		st.add_vertex(vertices[polygon[0]])
		mi.mesh = st.commit()
		mi.material_override = StandardMaterial3D.new()
		mi.material_override.albedo_color = CONFLICT_COLORS[color_index]
		mi.cast_shadow = GeometryInstance3D.SHADOW_CASTING_SETTING_OFF
		color_index = (color_index + 1) % CONFLICT_COLORS.size()
		polygon_meshes.push_back(mi)

func _get_point_key(vertex: Vector3, cell_size: Vector3) -> int:
	# Creates an int64 key for a vertex using the same logic as modules/navigation/nav_map.cpp.
	# See NavMap::get_point_key.
	var x: int = int(floor(vertex.x / cell_size.x)) & ((1 << 22) - 1)
	var y: int = int(floor(vertex.y / cell_size.y)) & ((1 << 23) - 1)
	var z: int = int(floor(vertex.z / cell_size.z)) & ((1 << 22) - 1)
	return x | (y << 21) | (z << 43)

func _get_edge_key(vertex1: Vector3, vertex2: Vector3, cell_size: Vector3) -> Array:
	# Similar to _get_point_key, this creates a 128 bit key for an edge using the same logic as
	# Godot's navigation module.
	var a: int = _get_point_key(vertex1, cell_size)
	var b: int = _get_point_key(vertex2, cell_size)
	if a < b:
		return [a, b]
	return [b, a]

func _postprocess_nav_mesh(nav_mesh: NavigationMesh) -> void:
	# Post-process the nav mesh to work around Godot issue #85548: nav mesh baking creates vertices
	# closer than the resolution of the navigation map allows. We round every vertex to the
	# nearest multiple of cell_size / cell_height, so that we can merge overly-close vertices.
	
	assert(nav_mesh != null)
	assert(nav_mesh.cell_size > 0.0)
	assert(nav_mesh.cell_height > 0.0)
	
	var cell_size: Vector3 = Vector3(nav_mesh.cell_size, nav_mesh.cell_height, nav_mesh.cell_size)
	
	var vertices: PackedVector3Array = nav_mesh.get_vertices()
	for i in range(vertices.size()):
		vertices[i] = (vertices[i] / cell_size).round() * cell_size
	
	var polygons: Array[PackedInt32Array] = []
	for i in range(nav_mesh.get_polygon_count()):
		var old_polygon: PackedInt32Array = nav_mesh.get_polygon(i)
		var new_polygon: PackedInt32Array = []
		
		# Remove duplicate vertices from the polygon:
		var polygon_vertices: PackedVector3Array = []
		for index in old_polygon:
			var vertex: Vector3 = vertices[index]
			if polygon_vertices.has(vertex):
				continue
			polygon_vertices.push_back(vertex)
			new_polygon.push_back(index)
		
		# If we removed some vertices, we might be remove the polygon too:
		if new_polygon.size() > 2:
			polygons.push_back(new_polygon)
	
	nav_mesh.clear_polygons()
	nav_mesh.set_vertices(vertices)
	for polygon in polygons:
		nav_mesh.add_polygon(polygon)

